/*******************************************************************************
 * Copyright (C) 2017 Philipp Spohr
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
package de.hhu.ba.yoshikoWrapper.tasks;

import java.awt.Window;
import java.util.Properties;

import javax.swing.SwingUtilities;

import org.cytoscape.application.swing.CytoPanel;
import org.cytoscape.application.swing.CytoPanelComponent;
import org.cytoscape.application.swing.CytoPanelName;
import org.cytoscape.application.swing.CytoPanelState;
import org.cytoscape.model.CyColumn;
import org.cytoscape.model.CyNode;
import org.cytoscape.work.AbstractTask;
import org.cytoscape.work.ContainsTunables;
import org.cytoscape.work.ObservableTask;
import org.cytoscape.work.TaskMonitor;
import org.cytoscape.work.TunableValidator;
import org.slf4j.Logger;

import de.hhu.ba.yoshikoWrapper.core.CyCore;
import de.hhu.ba.yoshikoWrapper.core.LocalizationManager;
import de.hhu.ba.yoshikoWrapper.core.NetworkParser;
import de.hhu.ba.yoshikoWrapper.core.ParameterSet;
import de.hhu.ba.yoshikoWrapper.core.StatusInformer;
import de.hhu.ba.yoshikoWrapper.cytoUtil.GraphAnalyzer;
import de.hhu.ba.yoshikoWrapper.cytoUtil.NodeMap;
import de.hhu.ba.yoshikoWrapper.graphModel.YoshikoCluster;
import de.hhu.ba.yoshikoWrapper.graphModel.YoshikoResult;
import de.hhu.ba.yoshikoWrapper.graphModel.YoshikoSolution;
import de.hhu.ba.yoshikoWrapper.logging.YoshikoLogger;
import de.hhu.ba.yoshikoWrapper.swig.ClusterEditingSolutions;
import de.hhu.ba.yoshikoWrapper.swig.CoreAlgorithm;
import de.hhu.ba.yoshikoWrapper.swig.IntVector;
import de.hhu.ba.yoshikoWrapper.swig.LibraryInput;
import de.hhu.ba.yoshikoWrapper.swig.LibraryInterface;
import de.hhu.ba.yoshikoWrapper.swing.components.ResultPanel;


public class AlgorithmTask extends AbstractTask implements ObservableTask, TunableValidator {

	//Constants
	private static final String SOLUTION_COLUMN_PREFIX = "yoshikoSolution_"; //TODO: Make customizable?

	//Symbolic links
	private static Logger logger = YoshikoLogger.getInstance().getLogger();


	private final Window statusWindow; //TODO: Make tunable?

	//Parameters

	/**The ParameterSet specifying how the algorithm is to be performed
	 *
	 */
	@ContainsTunables
	public ParameterSet parameterSet = null;



	//Temps, pointing to and need to be freed in C++
	private LibraryInput c_input;
	private ClusterEditingSolutions c_result;
	private CoreAlgorithm c_algorithm;

	private ResultPanel resultPanel;

	private YoshikoResult result;


	/**
	 * Default constructor, creates a new AlgorithmTask
	 * @param statusWindow The Window in which the status-bar is to be shown, can be null
	 * @param net The network that is to be clustered
	 * @param parameterSet The parameter set specifying the clustering mode
	 */
	public AlgorithmTask(
			Window statusWindow,
			ParameterSet parameterSet
			)
	{
		this.statusWindow = statusWindow;
		this.parameterSet = parameterSet;
	}

	@Override
	public void run(TaskMonitor taskMonitor) throws Exception {

		//TODO: Improve setProgress values

		taskMonitor.setTitle(LocalizationManager.get("yoshTask"));
		taskMonitor.setProgress(0.0);

		//Check current network
		if (parameterSet.net == null) {
			logger.warn("CoreAlgorithm called on a net that is NULL!");
			throw new Exception("CoreAlgorithm called on a net that is NULL!"); //TODO: Localize
		}

		//Set time limit
		LibraryInterface.setTimeLimit(parameterSet.timeLimit);
		if (parameterSet.disableMultiThreading) {
			LibraryInterface.setThreadLimit(1);
		}
		else {
			LibraryInterface.setThreadLimit(0);
		}

		//Create a node map to identify nodes and eges in the solution
		NodeMap nodeMap  = new NodeMap(parameterSet.net);
		taskMonitor.setProgress(0.1);

		//We identify the columns if they exist from their given names
		CyColumn weightColumn = parameterSet.weightColumnName != null ? parameterSet.net.getDefaultEdgeTable().getColumn(parameterSet.weightColumnName) : null;
		CyColumn permanentColumn = parameterSet.permanentColumnName != null ? parameterSet.net.getDefaultEdgeTable().getColumn(parameterSet.permanentColumnName) : null;
		CyColumn forbiddenColumn = parameterSet.forbiddenColumnName != null ? parameterSet.net.getDefaultEdgeTable().getColumn(parameterSet.forbiddenColumnName) : null;

		//Generate an c_input instance from the network
		c_input = NetworkParser.parseNetwork(
				parameterSet.net,
				nodeMap,
				weightColumn,
				permanentColumn,
				forbiddenColumn,
				parameterSet.defaultDeletionCost
				);
		taskMonitor.setProgress(0.2);

		boolean containsRealValues = GraphAnalyzer.containsRealValues(
				parameterSet.net,
				weightColumn,
				permanentColumn,
				forbiddenColumn,
				parameterSet.defaultInsertionCost,
				parameterSet.defaultDeletionCost);

		//Suggest reduction rules
		if (parameterSet.suggestReduction) {
			parameterSet.reductionRulesBitMask = GraphAnalyzer.suggestReductionRules(containsRealValues, parameterSet.useHeuristic);
		}

		//Set the default value for insertion cost
		c_input.setDefaultInsertionCost(parameterSet.defaultInsertionCost);

		//Call Yoshiko <<< Algorithm is performed here
		c_algorithm = LibraryInterface.getRun(c_input,
				parameterSet.solCount,
				parameterSet.reductionRulesBitMask,
				parameterSet.snrMultFactor,
				parameterSet.useHeuristic,
				parameterSet.usePartitionCuts,
				parameterSet.useTriangleCuts
		);


		c_algorithm.registerCplexInformer(new StatusInformer(statusWindow,taskMonitor));

		c_result  = c_algorithm.run();

		taskMonitor.setProgress(0.9);

		checkForInvalidSolutions(c_result);

		long numberOfSolutions = c_result.getNumberOfSolutions();

		taskMonitor.setStatusMessage("Found: "+numberOfSolutions+" solutions!"); //TODO: Localize

		result = new YoshikoResult(parameterSet.net,c_result.getFlags());

		//Loop over (multiple) solutions
		for (long i=0;i<numberOfSolutions;i++) {

			taskMonitor.setStatusMessage("Processing solution "+(i+1)+" of "+numberOfSolutions);

			YoshikoSolution solution = new YoshikoSolution(result,i);

			String columnName = SOLUTION_COLUMN_PREFIX+(i+1);

			parameterSet.net.getDefaultNodeTable().deleteColumn(columnName);
			parameterSet.net.getDefaultNodeTable().createColumn(columnName, String.class, false);

			//Fetch number of clusters in the solution
			long numberOfClusters = c_result.getNumberOfClusters(i);

			//Loop over clusters
			for (long k=0;k<numberOfClusters;k++) {
				//Create java instance
				YoshikoCluster cluster = new YoshikoCluster(solution,k);


				taskMonitor.setStatusMessage("Processing clusters "+(k+1)+" of "+numberOfClusters);

				IntVector clusterVector = c_result.getCluster(i, k);

				long sizeOfCluster = clusterVector.size();
				for (int l=0;l<sizeOfCluster;l++) { //Unsafe mismatch int long
					//taskMonitor.setStatusMessage("Processing node "+(l+1)+ " of "+sizeOfCluster);
					int nodeID = clusterVector.get(l);
					CyNode node = nodeMap.indexOf(nodeID); //<<< Another int/long conversion
					cluster.addNode(node);

					parameterSet.net.getRow(node).set(columnName, ""+(k+1)); //Add Cluster ID in table (Remove in final version?)
				}
				//Register clusters with solution for further reference
				solution.addCluster(cluster);
			}
			//Sort clusters by size, descending as the biggest clusters are usually the most relevant
			//solution.sortClusters(YoshikoCluster.lessThanComparator);
			//Register solution with c_result for further reference
			result.addSolution(solution);
		}

		taskMonitor.setProgress(0.99);
		//At this point the C++ objects can be freed, in case that the algorithm terminates beforehand they need to be cleaned-up elsewhere
		c_algorithm.delete();
		c_input.delete();
		c_result.delete();

		//Generate solutionsPanel
		resultPanel = new ResultPanel(result);

		//Show solution panel
		CyCore.registrar.registerService(resultPanel,CytoPanelComponent.class, new Properties());
		//Focus solution panel
		CytoPanel eastPanel = CyCore.swing.getCytoPanel(CytoPanelName.EAST);
		eastPanel.setSelectedIndex(eastPanel.indexOfComponent(resultPanel));
		//Show (might be invisible)
		eastPanel.setState(CytoPanelState.DOCK);
		//Blackest magic, attempts to scale the result panel to a good size
		SwingUtilities.invokeLater(
				new Runnable() {
					@Override
					public void run() {
						eastPanel.getThisComponent().setMinimumSize(eastPanel.getThisComponent().getPreferredSize());
						eastPanel.getThisComponent().revalidate();
					}
				}
			);
		//eastPanel.getThisComponent().revalidate();

	}

	private void checkForInvalidSolutions(ClusterEditingSolutions result) throws Exception {
		if (result == null) {
			//There was no c_result object generated at all
			throw new Exception(LocalizationManager.get("noFeasible"));
		}

		if (result.getNumberOfSolutions() == 0) {
			//The c_result object is empty
			throw new Exception(LocalizationManager.get("noFeasible"));
		}
		if (result.getNumberOfSolutions() == 1) {
			//The c_result doesn't contain any clusters
			if (result.getNumberOfClusters(0) == 0) {
				throw new Exception(LocalizationManager.get("noFeasible"));
			}
		}

	}

	@Override
	public void cancel() {
		//Free C++ resources
		if (c_algorithm != null) {
			System.out.println("Deleting CoreAlgorithm instance!");
			c_algorithm.cancel();
			c_algorithm.delete();
		}
		if (c_input != null) {
			c_input.delete();
		}
		if (c_result != null) {
			c_result.delete();
		}
		//Delete (incomplete) c_result panel
		if (resultPanel != null) {
			resultPanel.deleteResult();
		}
		super.cancel();
	}

	@SuppressWarnings("unchecked")
	@Override
	public <R> R getResults(Class<? extends R> type) {
		//We return the id of the result so we can work with the result from CMD
		if (type.equals(String.class)) {
			return (R) ("Result generated with ID: "+result.getID());
		}
		return null;
	}

	@Override
	public ValidationState getValidationState(Appendable errMsg) {
		//In order to validate the arguments for this task we simply check the ParameterSet
		return parameterSet.getValidationState(errMsg);
	}
}
